@@ -14,9 +14,7 @@
 #include <asm/unistd.h>
 
 #include "../fio.h"
-#include "../indirect.h"
-#include "../syslet.h"
-#include "../fls.h"
+#include "../lib/fls.h"
 
 #ifdef FIO_HAVE_SYSLET
 
@@ -224,7 +222,6 @@ static void fio_syslet_cleanup(struct thread_data *td)
 		free(sd->events);
 		free(sd->ring);
 		free(sd);
-		td->io_ops->data = NULL;
 	}
 }
 
@@ -245,7 +242,7 @@ static int fio_syslet_init(struct thread_data *td)
 	 */
 	ring_nr = td->o.iodepth;
 	if (ring_nr & (ring_nr - 1))
-		ring_nr = 1 << fls(ring_nr);
+		ring_nr = 1 << __fls(ring_nr);
 
 	ring_size = sizeof(struct syslet_ring) +
 			ring_nr * sizeof(struct syslet_completion);
