@@ -9,18 +9,20 @@
 
 #include "hash.h"
 
-static int clock_gettime_works = 0;
+static int clock_gettime_works;
+static struct timeval last_tv;
+static int last_tv_valid;
 
 #ifdef FIO_DEBUG_TIME
 
 #define HASH_BITS	8
 #define HASH_SIZE	(1 << HASH_BITS)
 
-static struct list_head hash[HASH_SIZE];
+static struct flist_head hash[HASH_SIZE];
 static int gtod_inited;
 
 struct gtod_log {
-	struct list_head list;
+	struct flist_head list;
 	void *caller;
 	unsigned long calls;
 };
@@ -28,10 +30,11 @@ struct gtod_log {
 static struct gtod_log *find_hash(void *caller)
 {
 	unsigned long h = hash_ptr(caller, HASH_BITS);
-	struct list_head *entry;
+	struct flist_head *entry;
 
-	list_for_each(entry, &hash[h]) {
-		struct gtod_log *log = list_entry(entry, struct gtod_log, list);
+	flist_for_each(entry, &hash[h]) {
+		struct gtod_log *log = flist_entry(entry, struct gtod_log,
+									list);
 
 		if (log->caller == caller)
 			return log;
@@ -48,12 +51,12 @@ static struct gtod_log *find_log(void *caller)
 		unsigned long h;
 
 		log = malloc(sizeof(*log));
-		INIT_LIST_HEAD(&log->list);
+		INIT_FLIST_HEAD(&log->list);
 		log->caller = caller;
 		log->calls = 0;
 
 		h = hash_ptr(caller, HASH_BITS);
-		list_add_tail(&log->list, &hash[h]);
+		flist_add_tail(&log->list, &hash[h]);
 	}
 
 	return log;
@@ -74,13 +77,14 @@ static void fio_exit fio_dump_gtod(void)
 	int i;
 
 	for (i = 0; i < HASH_SIZE; i++) {
-		struct list_head *entry;
+		struct flist_head *entry;
 		struct gtod_log *log;
 
-		list_for_each(entry, &hash[i]) {
-			log = list_entry(entry, struct gtod_log, list);
+		flist_for_each(entry, &hash[i]) {
+			log = flist_entry(entry, struct gtod_log, list);
 
-			printf("function %p, calls %lu\n", log->caller, log->calls);
+			printf("function %p, calls %lu\n", log->caller,
+								log->calls);
 			total_calls += log->calls;
 		}
 	}
@@ -93,7 +97,7 @@ static void fio_init gtod_init(void)
 	int i;
 
 	for (i = 0; i < HASH_SIZE; i++)
-		INIT_LIST_HEAD(&hash[i]);
+		INIT_FLIST_HEAD(&hash[i]);
 
 	gtod_inited = 1;
 }
@@ -126,4 +130,18 @@ gtod:
 		tp->tv_sec = ts.tv_sec;
 		tp->tv_usec = ts.tv_nsec / 1000;
 	}
+
+	/*
+	 * If Linux is using the tsc clock on non-synced processors,
+	 * sometimes time can appear to drift backwards. Fix that up.
+	 */
+	if (last_tv_valid) {
+		if (tp->tv_sec < last_tv.tv_sec)
+			tp->tv_sec = last_tv.tv_sec;
+		else if (last_tv.tv_sec == tp->tv_sec &&
+			 tp->tv_usec < last_tv.tv_usec)
+			tp->tv_usec = last_tv.tv_usec;
+	}
+	last_tv_valid = 1;
+	memcpy(&last_tv, tp, sizeof(*tp));
 }
