@@ -18,12 +18,13 @@
 
 #include "fio.h"
 
-static LIST_HEAD(engine_list);
+static FLIST_HEAD(engine_list);
 
 static int check_engine_ops(struct ioengine_ops *ops)
 {
 	if (ops->version != FIO_IOOPS_VERSION) {
-		log_err("bad ioops version %d (want %d)\n", ops->version, FIO_IOOPS_VERSION);
+		log_err("bad ioops version %d (want %d)\n", ops->version,
+							FIO_IOOPS_VERSION);
 		return 1;
 	}
 
@@ -37,7 +38,7 @@ static int check_engine_ops(struct ioengine_ops *ops)
 	 */
 	if (ops->flags & FIO_SYNCIO)
 		return 0;
-	
+
 	if (!ops->event) {
 		log_err("%s: no event handler\n", ops->name);
 		return 1;
@@ -50,31 +51,31 @@ static int check_engine_ops(struct ioengine_ops *ops)
 		log_err("%s: no queue handler\n", ops->name);
 		return 1;
 	}
-		
+
 	return 0;
 }
 
 void unregister_ioengine(struct ioengine_ops *ops)
 {
 	dprint(FD_IO, "ioengine %s unregistered\n", ops->name);
-	list_del(&ops->list);
-	INIT_LIST_HEAD(&ops->list);
+	flist_del(&ops->list);
+	INIT_FLIST_HEAD(&ops->list);
 }
 
 void register_ioengine(struct ioengine_ops *ops)
 {
 	dprint(FD_IO, "ioengine %s registered\n", ops->name);
-	INIT_LIST_HEAD(&ops->list);
-	list_add_tail(&ops->list, &engine_list);
+	INIT_FLIST_HEAD(&ops->list);
+	flist_add_tail(&ops->list, &engine_list);
 }
 
 static struct ioengine_ops *find_ioengine(const char *name)
 {
 	struct ioengine_ops *ops;
-	struct list_head *entry;
+	struct flist_head *entry;
 
-	list_for_each(entry, &engine_list) {
-		ops = list_entry(entry, struct ioengine_ops, list);
+	flist_for_each(entry, &engine_list) {
+		ops = flist_entry(entry, struct ioengine_ops, list);
 		if (!strcmp(name, ops->name))
 			return ops;
 	}
@@ -153,8 +154,10 @@ void close_ioengine(struct thread_data *td)
 {
 	dprint(FD_IO, "close ioengine %s\n", td->io_ops->name);
 
-	if (td->io_ops->cleanup)
+	if (td->io_ops->cleanup) {
 		td->io_ops->cleanup(td);
+		td->io_ops->data = NULL;
+	}
 
 	if (td->io_ops->dlhandle)
 		dlclose(td->io_ops->dlhandle);
@@ -168,10 +171,14 @@ int td_io_prep(struct thread_data *td, struct io_u *io_u)
 	dprint_io_u(io_u, "prep");
 	fio_ro_check(td, io_u);
 
+	lock_file(td, io_u->file, io_u->ddir);
+
 	if (td->io_ops->prep) {
 		int ret = td->io_ops->prep(td, io_u);
 
 		dprint(FD_IO, "->prep(%p)=%d\n", io_u, ret);
+		if (ret)
+			unlock_file(td, io_u->file);
 		return ret;
 	}
 
@@ -188,11 +195,17 @@ int td_io_getevents(struct thread_data *td, unsigned int min, unsigned int max,
 		if (r < 0)
 			goto out;
 	}
+	if (max > td->cur_depth)
+		max = td->cur_depth;
+	if (min > max)
+		max = min;
 
 	r = 0;
-	if (td->io_ops->getevents)
+	if (max && td->io_ops->getevents)
 		r = td->io_ops->getevents(td, min, max, t);
 out:
+	if (r >= 0)
+		io_u_mark_complete(td, r);
 	dprint(FD_IO, "getevents: %d\n", r);
 	return r;
 }
@@ -213,14 +226,15 @@ int td_io_queue(struct thread_data *td, struct io_u *io_u)
 	io_u->resid = 0;
 
 	if (td->io_ops->flags & FIO_SYNCIO) {
-		fio_gettime(&io_u->issue_time, NULL);
-		memcpy(&td->last_issue, &io_u->issue_time, sizeof(struct timeval));
+		if (fio_fill_issue_time(td))
+			fio_gettime(&io_u->issue_time, NULL);
 
 		/*
-		 * for a sync engine, set the timeout upfront
+		 * only used for iolog
 		 */
-		if (mtime_since(&td->timeout_end, &io_u->issue_time) < IO_U_TIMEOUT)
-			io_u_set_timeout(td);
+		if (td->o.read_iolog_file)
+			memcpy(&td->last_issue, &io_u->issue_time,
+					sizeof(struct timeval));
 	}
 
 	if (io_u->ddir != DDIR_SYNC)
@@ -228,14 +242,27 @@ int td_io_queue(struct thread_data *td, struct io_u *io_u)
 
 	ret = td->io_ops->queue(td, io_u);
 
-	if (ret != FIO_Q_BUSY)
-		io_u_mark_depth(td, io_u);
+	unlock_file(td, io_u->file);
 
-	if (ret == FIO_Q_QUEUED) {
+	if (!td->io_ops->commit) {
+		io_u_mark_submit(td, 1);
+		io_u_mark_complete(td, 1);
+	}
+
+	if (ret == FIO_Q_COMPLETED) {
+		if (io_u->ddir != DDIR_SYNC) {
+			io_u_mark_depth(td, 1);
+			td->ts.total_io_u[io_u->ddir]++;
+		}
+	} else if (ret == FIO_Q_QUEUED) {
 		int r;
 
-		td->io_u_queued++;
-		if (td->io_u_queued > td->o.iodepth_batch) {
+		if (io_u->ddir != DDIR_SYNC) {
+			td->io_u_queued++;
+			td->ts.total_io_u[io_u->ddir]++;
+		}
+
+		if (td->io_u_queued >= td->o.iodepth_batch) {
 			r = td_io_commit(td);
 			if (r < 0)
 				return r;
@@ -243,15 +270,15 @@ int td_io_queue(struct thread_data *td, struct io_u *io_u)
 	}
 
 	if ((td->io_ops->flags & FIO_SYNCIO) == 0) {
-		fio_gettime(&io_u->issue_time, NULL);
-		memcpy(&td->last_issue, &io_u->issue_time, sizeof(struct timeval));
+		if (fio_fill_issue_time(td))
+			fio_gettime(&io_u->issue_time, NULL);
 
 		/*
-		 * async engine, set the timeout here
+		 * only used for iolog
 		 */
-		if (ret == FIO_Q_QUEUED &&
-		    mtime_since(&td->timeout_end, &io_u->issue_time) < IO_U_TIMEOUT)
-			io_u_set_timeout(td);
+		if (td->o.read_iolog_file)
+			memcpy(&td->last_issue, &io_u->issue_time,
+					sizeof(struct timeval));
 	}
 
 	return ret;
@@ -263,8 +290,10 @@ int td_io_init(struct thread_data *td)
 
 	if (td->io_ops->init) {
 		ret = td->io_ops->init(td);
-		if (ret && td->o.iodepth > 1)
-			log_err("fio: io engine init failed. Perhaps try reducing io depth?\n");
+		if (ret && td->o.iodepth > 1) {
+			log_err("fio: io engine init failed. Perhaps try"
+				" reducing io depth?\n");
+		}
 	}
 
 	return ret;
@@ -274,10 +303,12 @@ int td_io_commit(struct thread_data *td)
 {
 	dprint(FD_IO, "calling ->commit(), depth %d\n", td->cur_depth);
 
-	if (!td->cur_depth)
+	if (!td->cur_depth || !td->io_u_queued)
 		return 0;
 
+	io_u_mark_depth(td, td->io_u_queued);
 	td->io_u_queued = 0;
+
 	if (td->io_ops->commit)
 		return td->io_ops->commit(td);
 
@@ -289,12 +320,22 @@ int td_io_open_file(struct thread_data *td, struct fio_file *f)
 	if (td->io_ops->open_file(td, f)) {
 		if (td->error == EINVAL && td->o.odirect)
 			log_err("fio: destination does not support O_DIRECT\n");
-		if (td->error == EMFILE)
-			log_err("fio: try reducing/setting openfiles (failed at %u of %u)\n", td->nr_open_files, td->o.nr_files);
+		if (td->error == EMFILE) {
+			log_err("fio: try reducing/setting openfiles (failed"
+				" at %u of %u)\n", td->nr_open_files,
+							td->o.nr_files);
+		}
 
 		return 1;
 	}
 
+	fio_file_reset(f);
+	f->flags |= FIO_FILE_OPEN;
+	f->flags &= ~FIO_FILE_CLOSING;
+
+	td->nr_open_files++;
+	get_file(f);
+
 	if (f->filetype == FIO_TYPE_PIPE) {
 		if (td_random(td)) {
 			log_err("fio: can't seek on pipes (no random io)\n");
@@ -302,12 +343,6 @@ int td_io_open_file(struct thread_data *td, struct fio_file *f)
 		}
 	}
 
-	f->last_free_lookup = 0;
-	f->last_completed_pos = 0;
-	f->last_pos = f->file_offset;
-	f->flags |= FIO_FILE_OPEN;
-	f->flags &= ~FIO_FILE_CLOSING;
-
 	if (td->io_ops->flags & FIO_DISKLESSIO)
 		goto done;
 
@@ -316,7 +351,6 @@ int td_io_open_file(struct thread_data *td, struct fio_file *f)
 
 	if (td->o.fadvise_hint &&
 	    (f->filetype == FIO_TYPE_BD || f->filetype == FIO_TYPE_FILE)) {
-		
 		int flags;
 
 		if (td_random(td))
@@ -331,12 +365,25 @@ int td_io_open_file(struct thread_data *td, struct fio_file *f)
 	}
 
 	if (f->file_map)
-		memset(f->file_map, 0, f->num_maps * sizeof(long));
+		memset(f->file_map, 0, f->num_maps * sizeof(int));
+
+#ifdef FIO_OS_DIRECTIO
+	/*
+	 * Some OS's have a distinct call to mark the file non-buffered,
+	 * instead of using O_DIRECT (Solaris)
+	 */
+	if (td->o.odirect) {
+		int ret = fio_set_odirect(f->fd);
+
+		if (ret) {
+			td_verror(td, ret, "fio_set_odirect");
+			goto err;
+		}
+	}
+#endif
 
 done:
 	log_file(td, f, FIO_LOG_OPEN_FILE);
-	td->nr_open_files++;
-	get_file(f);
 	return 0;
 err:
 	if (td->io_ops->close_file)
@@ -344,7 +391,7 @@ err:
 	return 1;
 }
 
-void td_io_close_file(struct thread_data *td, struct fio_file *f)
+int td_io_close_file(struct thread_data *td, struct fio_file *f)
 {
 	if (!(f->flags & FIO_FILE_CLOSING))
 		log_file(td, f, FIO_LOG_CLOSE_FILE);
@@ -354,5 +401,7 @@ void td_io_close_file(struct thread_data *td, struct fio_file *f)
 	 */
 	f->flags |= FIO_FILE_CLOSING;
 
-	put_file(td, f);
+	unlock_file_all(td, f);
+
+	return put_file(td, f);
 }
