@@ -5,10 +5,14 @@
 #include <string.h>
 #include <getopt.h>
 #include <assert.h>
+#include <libgen.h>
+#include <fcntl.h>
+#include <sys/types.h>
+#include <sys/stat.h>
 
 #include "fio.h"
 #include "parse.h"
-#include "fls.h"
+#include "lib/fls.h"
 
 #define td_var_offset(var)	((size_t) &((struct thread_options *)0)->var)
 
@@ -66,7 +70,9 @@ static int str_bssplit_cb(void *data, const char *input)
 		 */
 		if (i == td->o.bssplit_nr) {
 			td->o.bssplit_nr <<= 1;
-			td->o.bssplit = realloc(td->o.bssplit, td->o.bssplit_nr * sizeof(struct bssplit));
+			td->o.bssplit = realloc(td->o.bssplit,
+						td->o.bssplit_nr
+						  * sizeof(struct bssplit));
 		}
 
 		perc_str = strstr(fname, "/");
@@ -148,8 +154,10 @@ static int str_rw_cb(void *data, const char *str)
 	char *nr = get_opt_postfix(str);
 
 	td->o.ddir_nr = 1;
-	if (nr)
+	if (nr) {
 		td->o.ddir_nr = atoi(nr);
+		free(nr);
+	}
 
 	return 0;
 }
@@ -175,6 +183,24 @@ static int str_lockmem_cb(void fio_unused *data, unsigned long *val)
 	return 0;
 }
 
+static int str_rwmix_read_cb(void *data, unsigned int *val)
+{
+	struct thread_data *td = data;
+
+	td->o.rwmix[DDIR_READ] = *val;
+	td->o.rwmix[DDIR_WRITE] = 100 - *val;
+	return 0;
+}
+
+static int str_rwmix_write_cb(void *data, unsigned int *val)
+{
+	struct thread_data *td = data;
+
+	td->o.rwmix[DDIR_WRITE] = *val;
+	td->o.rwmix[DDIR_READ] = 100 - *val;
+	return 0;
+}
+
 #ifdef FIO_HAVE_IOPRIO
 static int str_prioclass_cb(void *data, unsigned int *val)
 {
@@ -261,8 +287,49 @@ static int str_fst_cb(void *data, const char *str)
 	char *nr = get_opt_postfix(str);
 
 	td->file_service_nr = 1;
-	if (nr)
+	if (nr) {
 		td->file_service_nr = atoi(nr);
+		free(nr);
+	}
+
+	return 0;
+}
+
+static int check_dir(struct thread_data *td, char *fname)
+{
+	char file[PATH_MAX], *dir;
+	int elen = 0;
+
+	if (td->o.directory) {
+		strcpy(file, td->o.directory);
+		strcat(file, "/");
+		elen = strlen(file);
+	}
+
+	sprintf(file + elen, "%s", fname);
+	dir = dirname(file);
+
+#if 0
+	{
+	struct stat sb;
+	/*
+	 * We can't do this on FIO_DISKLESSIO engines. The engine isn't loaded
+	 * yet, so we can't do this check right here...
+	 */
+	if (lstat(dir, &sb) < 0) {
+		int ret = errno;
+
+		log_err("fio: %s is not a directory\n", dir);
+		td_verror(td, ret, "lstat");
+		return 1;
+	}
+
+	if (!S_ISDIR(sb.st_mode)) {
+		log_err("fio: %s is not a directory\n", dir);
+		return 1;
+	}
+	}
+#endif
 
 	return 0;
 }
@@ -283,6 +350,10 @@ static int str_filename_cb(void *data, const char *input)
 	while ((fname = strsep(&str, ":")) != NULL) {
 		if (!strlen(fname))
 			break;
+		if (check_dir(td, fname)) {
+			free(p);
+			return 1;
+		}
 		add_file(td, fname);
 		td->o.nr_files++;
 	}
@@ -297,8 +368,10 @@ static int str_directory_cb(void *data, const char fio_unused *str)
 	struct stat sb;
 
 	if (lstat(td->o.directory, &sb) < 0) {
+		int ret = errno;
+
 		log_err("fio: %s is not a directory\n", td->o.directory);
-		td_verror(td, errno, "lstat");
+		td_verror(td, ret, "lstat");
 		return 1;
 	}
 	if (!S_ISDIR(sb.st_mode)) {
@@ -337,7 +410,7 @@ static int str_verify_pattern_cb(void *data, unsigned int *off)
 	struct thread_data *td = data;
 	unsigned int msb;
 
-	msb = fls(*off);
+	msb = __fls(*off);
 	if (msb <= 8)
 		td->o.verify_pattern_bytes = 1;
 	else if (msb <= 16)
@@ -351,6 +424,34 @@ static int str_verify_pattern_cb(void *data, unsigned int *off)
 	return 0;
 }
 
+static int str_lockfile_cb(void *data, const char *str)
+{
+	struct thread_data *td = data;
+	char *nr = get_opt_postfix(str);
+
+	td->o.lockfile_batch = 1;
+	if (nr) {
+		td->o.lockfile_batch = atoi(nr);
+		free(nr);
+	}
+
+	return 0;
+}
+
+static int str_gtod_reduce_cb(void *data, int *il)
+{
+	struct thread_data *td = data;
+	int val = *il;
+
+	td->o.disable_clat = !!val;
+	td->o.disable_slat = !!val;
+	td->o.disable_bw = !!val;
+	if (val)
+		td->tv_cache_mask = 63;
+
+	return 0;
+}
+
 #define __stringify_1(x)	#x
 #define __stringify(x)		__stringify_1(x)
 
@@ -382,8 +483,33 @@ static struct fio_option options[] = {
 		.type	= FIO_OPT_STR_STORE,
 		.off1	= td_var_offset(filename),
 		.cb	= str_filename_cb,
+		.prio	= 1, /* must come before "directory" */
 		.help	= "File(s) to use for the workload",
 	},
+	{
+		.name	= "lockfile",
+		.type	= FIO_OPT_STR,
+		.cb	= str_lockfile_cb,
+		.off1	= td_var_offset(file_lock_mode),
+		.help	= "Lock file when doing IO to it",
+		.parent	= "filename",
+		.def	= "none",
+		.posval = {
+			  { .ival = "none",
+			    .oval = FILE_LOCK_NONE,
+			    .help = "No file locking",
+			  },
+			  { .ival = "exclusive",
+			    .oval = FILE_LOCK_EXCLUSIVE,
+			    .help = "Exclusive file lock",
+			  },
+			  {
+			    .ival = "readwrite",
+			    .oval = FILE_LOCK_READWRITE,
+			    .help = "Read vs write lock",
+			  },
+		},
+	},
 	{
 		.name	= "opendir",
 		.type	= FIO_OPT_STR_STORE,
@@ -451,6 +577,11 @@ static struct fio_option options[] = {
 			  { .ival = "posixaio",
 			    .help = "POSIX asynchronous IO",
 			  },
+#endif
+#ifdef FIO_HAVE_SOLARISAIO
+			  { .ival = "solarisaio",
+			    .help = "Solaris native asynchronous IO",
+			  },
 #endif
 			  { .ival = "mmap",
 			    .help = "Memory mapped IO",
@@ -502,6 +633,7 @@ static struct fio_option options[] = {
 	},
 	{
 		.name	= "iodepth_batch",
+		.alias	= "iodepth_batch_submit",
 		.type	= FIO_OPT_INT,
 		.off1	= td_var_offset(iodepth_batch),
 		.help	= "Number of IO to submit in one go",
@@ -509,6 +641,15 @@ static struct fio_option options[] = {
 		.minval	= 1,
 		.def	= "1",
 	},
+	{
+		.name	= "iodepth_batch_complete",
+		.type	= FIO_OPT_INT,
+		.off1	= td_var_offset(iodepth_batch_complete),
+		.help	= "Number of IO to retrieve in one go",
+		.parent	= "iodepth",
+		.minval	= 0,
+		.def	= "1",
+	},
 	{
 		.name	= "iodepth_low",
 		.type	= FIO_OPT_INT,
@@ -599,6 +740,14 @@ static struct fio_option options[] = {
 		.help	= "Accept potential duplicate random blocks",
 		.parent = "rw",
 	},
+	{
+		.name	= "softrandommap",
+		.type	= FIO_OPT_BOOL,
+		.off1	= td_var_offset(softrandommap),
+		.help	= "Set norandommap if randommap allocation fails",
+		.parent	= "norandommap",
+		.def	= "0",
+	},
 	{
 		.name	= "nrfiles",
 		.type	= FIO_OPT_INT,
@@ -702,6 +851,12 @@ static struct fio_option options[] = {
 		.off1	= td_var_offset(time_based),
 		.help	= "Keep running until runtime/timeout is met",
 	},
+	{
+		.name	= "ramp_time",
+		.type	= FIO_OPT_STR_VAL_TIME,
+		.off1	= td_var_offset(ramp_time),
+		.help	= "Ramp up time before measuring performance",
+	},
 	{
 		.name	= "mem",
 		.alias	= "iomem",
@@ -760,6 +915,14 @@ static struct fio_option options[] = {
 			    .oval = VERIFY_CRC32,
 			    .help = "Use crc32 checksums for verification",
 			  },
+			  { .ival = "crc32c-intel",
+			    .oval = VERIFY_CRC32C_INTEL,
+			    .help = "Use hw crc32c checksums for verification",
+			  },
+			  { .ival = "crc32c",
+			    .oval = VERIFY_CRC32C,
+			    .help = "Use crc32c checksums for verification",
+			  },
 			  { .ival = "crc16",
 			    .oval = VERIFY_CRC16,
 			    .help = "Use crc16 checksums for verification",
@@ -816,7 +979,7 @@ static struct fio_option options[] = {
 		.type	= FIO_OPT_STR_VAL_INT,
 		.help	= "Offset verify header location by N bytes",
 		.def	= "0",
-		.cb	= str_verify_offset_cb,	
+		.cb	= str_verify_offset_cb,
 		.parent	= "verify",
 	},
 	{
@@ -890,7 +1053,7 @@ static struct fio_option options[] = {
 	{
 		.name	= "rwmixread",
 		.type	= FIO_OPT_INT,
-		.off1	= td_var_offset(rwmix[DDIR_READ]),
+		.cb	= str_rwmix_read_cb,
 		.maxval	= 100,
 		.help	= "Percentage of mixed workload that is reads",
 		.def	= "50",
@@ -898,18 +1061,14 @@ static struct fio_option options[] = {
 	{
 		.name	= "rwmixwrite",
 		.type	= FIO_OPT_INT,
-		.off1	= td_var_offset(rwmix[DDIR_WRITE]),
+		.cb	= str_rwmix_write_cb,
 		.maxval	= 100,
 		.help	= "Percentage of mixed workload that is writes",
 		.def	= "50",
 	},
 	{
 		.name	= "rwmixcycle",
-		.type	= FIO_OPT_INT,
-		.off1	= td_var_offset(rwmixcycle),
-		.help	= "Cycle period for mixed read/write workloads (msec)",
-		.def	= "500",
-		.parent	= "rwmixread",
+		.type	= FIO_OPT_DEPRECATED,
 	},
 	{
 		.name	= "nice",
@@ -1014,7 +1173,8 @@ static struct fio_option options[] = {
 		.name	= "bwavgtime",
 		.type	= FIO_OPT_INT,
 		.off1	= td_var_offset(bw_avg_time),
-		.help	= "Time window over which to calculate bandwidth (msec)",
+		.help	= "Time window over which to calculate bandwidth"
+			  " (msec)",
 		.def	= "500",
 	},
 	{
@@ -1118,7 +1278,7 @@ static struct fio_option options[] = {
 	},
 	{
 		.name	= "hugepage-size",
-		.type	= FIO_OPT_STR_VAL,
+		.type	= FIO_OPT_STR_VAL_INT,
 		.off1	= td_var_offset(hugepage_size),
 		.help	= "When using hugepages, specify size of each page",
 		.def	= __stringify(FIO_HUGE_PAGE),
@@ -1135,15 +1295,52 @@ static struct fio_option options[] = {
 		.off1	= td_var_offset(zero_buffers),
 		.help	= "Init IO buffers to all zeroes",
 	},
+	{
+		.name	= "refill_buffers",
+		.type	= FIO_OPT_STR_SET,
+		.off1	= td_var_offset(refill_buffers),
+		.help	= "Refill IO buffers on every IO submit",
+	},
 #ifdef FIO_HAVE_DISK_UTIL
 	{
 		.name	= "disk_util",
 		.type	= FIO_OPT_BOOL,
 		.off1	= td_var_offset(do_disk_util),
-		.help	= "Log disk utilization stats",
+		.help	= "Log disk utilization statistics",
 		.def	= "1",
 	},
 #endif
+	{
+		.name	= "gtod_reduce",
+		.type	= FIO_OPT_BOOL,
+		.help	= "Greatly reduce number of gettimeofday() calls",
+		.cb	= str_gtod_reduce_cb,
+		.def	= "0",
+	},
+	{
+		.name	= "disable_clat",
+		.type	= FIO_OPT_BOOL,
+		.off1	= td_var_offset(disable_clat),
+		.help	= "Disable completion latency numbers",
+		.parent	= "gtod_reduce",
+		.def	= "0",
+	},
+	{
+		.name	= "disable_slat",
+		.type	= FIO_OPT_BOOL,
+		.off1	= td_var_offset(disable_slat),
+		.help	= "Disable submissionn latency numbers",
+		.parent	= "gtod_reduce",
+		.def	= "0",
+	},
+	{
+		.name	= "disable_bw_measurement",
+		.type	= FIO_OPT_BOOL,
+		.off1	= td_var_offset(disable_bw),
+		.help	= "Disable bandwidth logging",
+		.parent	= "gtod_reduce",
+		.def	= "0",
+	},
 	{
 		.name = NULL,
 	},
@@ -1162,7 +1359,7 @@ void fio_options_dup_and_init(struct option *long_options)
 
 	o = &options[0];
 	while (o->name) {
-		long_options[i].name = o->name;
+		long_options[i].name = (char *) o->name;
 		long_options[i].val = FIO_GETOPT_JOB;
 		if (o->type == FIO_OPT_STR_SET)
 			long_options[i].has_arg = no_argument;
@@ -1175,9 +1372,16 @@ void fio_options_dup_and_init(struct option *long_options)
 	}
 }
 
-int fio_option_parse(struct thread_data *td, const char *opt)
+int fio_options_parse(struct thread_data *td, char **opts, int num_opts)
 {
-	return parse_option(opt, options, td);
+	int i, ret;
+
+	sort_options(opts, options, num_opts);
+
+	for (ret = 0, i = 0; i < num_opts; i++)
+		ret |= parse_option(opts[i], options, td);
+
+	return ret;
 }
 
 int fio_cmd_option_parse(struct thread_data *td, const char *opt, char *val)
