@@ -8,8 +8,12 @@
 #include <string.h>
 #include <errno.h>
 #include <limits.h>
+#include <stdlib.h>
 
 #include "parse.h"
+#include "debug.h"
+
+static struct fio_option *fio_options;
 
 static int vp_cmp(const void *p1, const void *p2)
 {
@@ -69,37 +73,37 @@ static void show_option_values(struct fio_option *o)
 static unsigned long get_mult_time(char c)
 {
 	switch (c) {
-		case 'm':
-		case 'M':
-			return 60;
-		case 'h':
-		case 'H':
-			return 60 * 60;
-		case 'd':
-		case 'D':
-			return 24 * 60 * 60;
-		default:
-			return 1;
+	case 'm':
+	case 'M':
+		return 60;
+	case 'h':
+	case 'H':
+		return 60 * 60;
+	case 'd':
+	case 'D':
+		return 24 * 60 * 60;
+	default:
+		return 1;
 	}
 }
 
 static unsigned long get_mult_bytes(char c)
 {
 	switch (c) {
-		case 'k':
-		case 'K':
-			return 1024;
-		case 'm':
-		case 'M':
-			return 1024 * 1024;
-		case 'g':
-		case 'G':
-			return 1024 * 1024 * 1024;
-		case 'e':
-		case 'E':
-			return 1024 * 1024 * 1024 * 1024UL;
-		default:
-			return 1;
+	case 'k':
+	case 'K':
+		return 1024;
+	case 'm':
+	case 'M':
+		return 1024 * 1024;
+	case 'g':
+	case 'G':
+		return 1024 * 1024 * 1024;
+	case 'e':
+	case 'E':
+		return 1024 * 1024 * 1024 * 1024UL;
+	default:
+		return 1;
 	}
 }
 
@@ -229,6 +233,9 @@ static int __handle_option(struct fio_option *o, const char *ptr, void *data,
 	char **cp;
 	int ret = 0, is_time = 0;
 
+	dprint(FD_PARSE, "__handle_option=%s, type=%d, ptr=%s\n", o->name,
+							o->type, ptr);
+
 	if (!ptr && o->type != FIO_OPT_STR_SET) {
 		fprintf(stderr, "Option %s requires an argument\n", o->name);
 		return 1;
@@ -278,11 +285,13 @@ static int __handle_option(struct fio_option *o, const char *ptr, void *data,
 			break;
 
 		if (o->maxval && ull > o->maxval) {
-			fprintf(stderr, "max value out of range: %lld (%d max)\n", ull, o->maxval);
+			fprintf(stderr, "max value out of range: %lld"
+					" (%d max)\n", ull, o->maxval);
 			return 1;
 		}
 		if (o->minval && ull < o->minval) {
-			fprintf(stderr, "min value out of range: %lld (%d min)\n", ull, o->minval);
+			fprintf(stderr, "min value out of range: %lld"
+					" (%d min)\n", ull, o->minval);
 			return 1;
 		}
 
@@ -337,7 +346,8 @@ static int __handle_option(struct fio_option *o, const char *ptr, void *data,
 		p1 = tmp;
 
 		ret = 1;
-		if (!check_range_bytes(p1, &ul1) && !check_range_bytes(p2, &ul2)) {
+		if (!check_range_bytes(p1, &ul1) &&
+		    !check_range_bytes(p2, &ul2)) {
 			ret = 0;
 			if (ul1 > ul2) {
 				unsigned long foo = ul1;
@@ -367,11 +377,13 @@ static int __handle_option(struct fio_option *o, const char *ptr, void *data,
 			break;
 
 		if (o->maxval && il > (int) o->maxval) {
-			fprintf(stderr, "max value out of range: %d (%d max)\n", il, o->maxval);
+			fprintf(stderr, "max value out of range: %d (%d max)\n",
+								il, o->maxval);
 			return 1;
 		}
 		if (o->minval && il < o->minval) {
-			fprintf(stderr, "min value out of range: %d (%d min)\n", il, o->minval);
+			fprintf(stderr, "min value out of range: %d (%d min)\n",
+								il, o->minval);
 			return 1;
 		}
 
@@ -401,6 +413,9 @@ static int __handle_option(struct fio_option *o, const char *ptr, void *data,
 		}
 		break;
 	}
+	case FIO_OPT_DEPRECATED:
+		fprintf(stdout, "Option %s is deprecated\n", o->name);
+		break;
 	default:
 		fprintf(stderr, "Bad option type %u\n", o->type);
 		ret = 1;
@@ -414,6 +429,8 @@ static int handle_option(struct fio_option *o, const char *ptr, void *data)
 	const char *ptr2 = NULL;
 	int r1, r2;
 
+	dprint(FD_PARSE, "handle_option=%s, ptr=%s\n", o->name, ptr);
+
 	/*
 	 * See if we have a second set of parameters, hidden after a comma.
 	 * Do this before parsing the first round, to check if we should
@@ -445,6 +462,56 @@ static int handle_option(struct fio_option *o, const char *ptr, void *data)
 	return r1 && r2;
 }
 
+static struct fio_option *get_option(const char *opt,
+				     struct fio_option *options, char **post)
+{
+	struct fio_option *o;
+	char *ret;
+
+	ret = strchr(opt, '=');
+	if (ret) {
+		*post = ret;
+		*ret = '\0';
+		ret = (char *) opt;
+		(*post)++;
+		o = find_option(options, ret);
+	} else {
+		o = find_option(options, opt);
+		*post = NULL;
+	}
+
+	return o;
+}
+
+static int opt_cmp(const void *p1, const void *p2)
+{
+	struct fio_option *o1, *o2;
+	char *s1, *s2, *foo;
+	int ret;
+
+	s1 = strdup(*((char **) p1));
+	s2 = strdup(*((char **) p2));
+
+	o1 = get_option(s1, fio_options, &foo);
+	o2 = get_option(s2, fio_options, &foo);
+
+	if ((!o1 && o2) || (o1 && !o2))
+		ret = 0;
+	else
+		ret = o2->prio - o1->prio;
+
+	free(s1);
+	free(s2);
+	return ret;
+}
+
+void sort_options(char **opts, struct fio_option *options, int num_opts)
+{
+	fio_options = options;
+	qsort(opts, num_opts, sizeof(char *), opt_cmp);
+	fio_options = NULL;
+}
+
 int parse_cmd_option(const char *opt, const char *val,
 		     struct fio_option *options, void *data)
 {
@@ -463,26 +530,63 @@ int parse_cmd_option(const char *opt, const char *val,
 	return 1;
 }
 
+/*
+ * Return a copy of the input string with substrings of the form ${VARNAME}
+ * substituted with the value of the environment variable VARNAME.  The
+ * substitution always occurs, even if VARNAME is empty or the corresponding
+ * environment variable undefined.
+ */
+static char *option_dup_subs(const char *opt)
+{
+	char out[OPT_LEN_MAX+1];
+	char in[OPT_LEN_MAX+1];
+	char *outptr = out;
+	char *inptr = in;
+	char *ch1, *ch2, *env;
+	ssize_t nchr = OPT_LEN_MAX;
+	size_t envlen;
+
+	in[OPT_LEN_MAX] = '\0';
+	strncpy(in, opt, OPT_LEN_MAX);
+
+	while (*inptr && nchr > 0) {
+		if (inptr[0] == '$' && inptr[1] == '{') {
+			ch2 = strchr(inptr, '}');
+			if (ch2 && inptr+1 < ch2) {
+				ch1 = inptr+2;
+				inptr = ch2+1;
+				*ch2 = '\0';
+
+				env = getenv(ch1);
+				if (env) {
+					envlen = strlen(env);
+					if (envlen <= nchr) {
+						memcpy(outptr, env, envlen);
+						outptr += envlen;
+						nchr -= envlen;
+					}
+				}
+
+				continue;
+			}
+		}
+
+		*outptr++ = *inptr++;
+		--nchr;
+	}
+
+	*outptr = '\0';
+	return strdup(out);
+}
+
 int parse_option(const char *opt, struct fio_option *options, void *data)
 {
 	struct fio_option *o;
-	char *pre, *post;
-	char *tmp;
+	char *post, *tmp;
 
-	tmp = strdup(opt);
-
-	pre = strchr(tmp, '=');
-	if (pre) {
-		post = pre;
-		*pre = '\0';
-		pre = tmp;
-		post++;
-		o = find_option(options, pre);
-	} else {
-		o = find_option(options, tmp);
-		post = NULL;
-	}
+	tmp = option_dup_subs(opt);
 
+	o = get_option(tmp, options, &post);
 	if (!o) {
 		fprintf(stderr, "Bad option %s\n", tmp);
 		free(tmp);
@@ -573,24 +677,49 @@ static struct fio_option *find_child(struct fio_option *options,
 	return NULL;
 }
 
-static void print_option(struct fio_option *o, struct fio_option *org,
-			 int level)
+static void __print_option(struct fio_option *o, struct fio_option *org,
+			   int level)
 {
 	char name[256], *p;
+	int depth;
 
 	if (!o)
 		return;
 	if (!org)
 		org = o;
-	
+
 	p = name;
-	if (level)
-		p += sprintf(p, "%s", "    ");
+	depth = level;
+	while (depth--)
+		p += sprintf(p, "%s", "  ");
 
 	sprintf(p, "%s", o->name);
 
 	printf("%-24s: %s\n", name, o->help);
-	print_option(find_child(o, org), org, level + 1);
+}
+
+static void print_option(struct fio_option *o)
+{
+	struct fio_option *parent;
+	struct fio_option *__o;
+	unsigned int printed;
+	unsigned int level;
+
+	__print_option(o, NULL, 0);
+	parent = o;
+	level = 0;
+	do {
+		level++;
+		printed = 0;
+
+		while ((__o = find_child(o, parent)) != NULL) {
+			__print_option(__o, o, level);
+			o = __o;
+			printed++;
+		}
+
+		parent = o;
+	} while (printed);
 }
 
 int show_cmd_help(struct fio_option *options, const char *name)
@@ -608,6 +737,9 @@ int show_cmd_help(struct fio_option *options, const char *name)
 	for (o = &options[0]; o->name; o++) {
 		int match = 0;
 
+		if (o->type == FIO_OPT_DEPRECATED)
+			continue;
+
 		if (name) {
 			if (!strcmp(name, o->name) ||
 			    (o->alias && !strcmp(name, o->alias)))
@@ -629,7 +761,7 @@ int show_cmd_help(struct fio_option *options, const char *name)
 				printf("%24s: %s\n", o->name, o->help);
 			if (show_all) {
 				if (!o->parent)
-					print_option(o, NULL, 0);
+					print_option(o);
 				continue;
 			}
 		}
@@ -661,6 +793,8 @@ void fill_default_options(void *data, struct fio_option *options)
 {
 	struct fio_option *o;
 
+	dprint(FD_PARSE, "filling default options\n");
+
 	for (o = &options[0]; o->name; o++)
 		if (o->def)
 			handle_option(o, o->def, data);
@@ -674,18 +808,29 @@ void options_init(struct fio_option *options)
 {
 	struct fio_option *o;
 
+	dprint(FD_PARSE, "init options\n");
+
 	for (o = &options[0]; o->name; o++) {
+		if (o->type == FIO_OPT_DEPRECATED)
+			continue;
 		if (o->type == FIO_OPT_BOOL) {
 			o->minval = 0;
 			o->maxval = 1;
 		}
-		if (o->type == FIO_OPT_STR_SET && o->def)
-			fprintf(stderr, "Option %s: string set option with default will always be true\n", o->name);
-		if (!o->cb && !o->off1)
-			fprintf(stderr, "Option %s: neither cb nor offset given\n", o->name);
+		if (o->type == FIO_OPT_STR_SET && o->def) {
+			fprintf(stderr, "Option %s: string set option with"
+					" default will always be true\n",
+						o->name);
+		}
+		if (!o->cb && !o->off1) {
+			fprintf(stderr, "Option %s: neither cb nor offset"
+					" given\n", o->name);
+		}
 		if (o->type == FIO_OPT_STR || o->type == FIO_OPT_STR_STORE)
 			continue;
-		if (o->cb && (o->off1 || o->off2 || o->off3 || o->off4))
-			fprintf(stderr, "Option %s: both cb and offset given\n", o->name);
+		if (o->cb && (o->off1 || o->off2 || o->off3 || o->off4)) {
+			fprintf(stderr, "Option %s: both cb and offset given\n",
+								 o->name);
+		}
 	}
 }
