@@ -6,9 +6,10 @@
 static struct timeval genesis;
 static unsigned long ns_granularity;
 
-unsigned long utime_since(struct timeval *s, struct timeval *e)
+unsigned long long utime_since(struct timeval *s, struct timeval *e)
 {
 	long sec, usec;
+	unsigned long long ret;
 
 	sec = e->tv_sec - s->tv_sec;
 	usec = e->tv_usec - s->tv_usec;
@@ -17,12 +18,18 @@ unsigned long utime_since(struct timeval *s, struct timeval *e)
 		usec += 1000000;
 	}
 
-	sec *= 1000000UL;
+	/*
+	 * time warp bug on some kernels?
+	 */
+	if (sec < 0 || (sec == 0 && usec < 0))
+		return 0;
 
-	return sec + usec;
+	ret = sec * 1000000ULL + usec;
+
+	return ret;
 }
 
-unsigned long utime_since_now(struct timeval *s)
+unsigned long long utime_since_now(struct timeval *s)
 {
 	struct timeval t;
 
@@ -32,7 +39,7 @@ unsigned long utime_since_now(struct timeval *s)
 
 unsigned long mtime_since(struct timeval *s, struct timeval *e)
 {
-	long sec, usec;
+	long sec, usec, ret;
 
 	sec = e->tv_sec - s->tv_sec;
 	usec = e->tv_usec - s->tv_usec;
@@ -43,8 +50,15 @@ unsigned long mtime_since(struct timeval *s, struct timeval *e)
 
 	sec *= 1000UL;
 	usec /= 1000UL;
+	ret = sec + usec;
 
-	return sec + usec;
+	/*
+	 * time warp bug on some kernels?
+	 */
+	if (ret < 0)
+		ret = 0;
+
+	return ret;
 }
 
 unsigned long mtime_since_now(struct timeval *s)
@@ -150,6 +164,29 @@ unsigned long mtime_since_genesis(void)
 	return mtime_since_now(&genesis);
 }
 
+int in_ramp_time(struct thread_data *td)
+{
+	return td->o.ramp_time && !td->ramp_time_over;
+}
+
+int ramp_time_over(struct thread_data *td)
+{
+	struct timeval tv;
+
+	if (!td->o.ramp_time || td->ramp_time_over)
+		return 1;
+
+	fio_gettime(&tv, NULL);
+	if (mtime_since(&td->epoch, &tv) >= td->o.ramp_time * 1000) {
+		td->ramp_time_over = 1;
+		reset_all_stats(td);
+		td_set_runstate(td, TD_RAMP);
+		return 1;
+	}
+
+	return 0;
+}
+
 static void fio_init time_init(void)
 {
 	int i;
